#ifndef C99TREE_TREE_H
#define C99TREE_TREE_H
typedef struct c99c_node_s c99c_node_t;

#ifndef C99TREE_CC_H
#error do not include tree.h directly, include c99tree-cc.h
#endif

#include <porty/c99tree-cc/location.h>

/* tree node types */
typedef enum c99c_node_type_e {
	T_CALL,          /* function call instruction */
	T_STATEMENT,
	T_TYPE,          /* type container for qualifiers, specifier, typenames, attribs */
	T_DECL,
	T_FUNCDEF,
	T_ARGLIST,       /* function arguments */
	T_ARG,           /* a specific argument */
	T_BLOCK,         /* compound statement */
	T_CLASS,
	T_TYPEOF,
	T_ATTRLIST,
	T_ATTR,
	T_PTR,
	T_INDEX,
	T_QUALIFIER,
	T_ID,
	T_ELLIPSIS,      /* ... */
	T_ENUM,
	T_INIT,
	T_STRUCT,
	T_BITS,          /* :X part of a bitfield */
	T_ASM,
	T_LABEL,         /* jump and switch labels */
	T_CONDITION,     /* condition section of while(), for() and if() */
	T_THEN,
	T_ELSE,
	T_OP,            /* operator */
	T_ICONST,        /* integer constant */
	T_FCONST,        /* floating point constant */
	T_SCONST,        /* string constant */
	T_SCONSTLIST,    /* appended string constants */
	T_DEBUG,         /* dummy node for debugging - shouldn't show up in any of the production output */
	T_UNION,
	T_PRAGMA,
	T_TYPECUSTOM,    /* typedef'd types */
	T_TYPESPEC,      /* long, short, signed, unsigned, _Complex and _Imaginary */
	T_TYPENAME,      /* name of the type (like int, float, etc.) */
	T_TYPEDEF,
	T_EXPRLIST,      /* comma separated list of expressions (1,2,3) */
	T_EXPRBRACKET,   /* a bracketed expression, i.e. (expr) */
	T_CSE,           /* compount expression statement "({ expr })" */
	T_COMMENT,       /* goes in the token tree */
	T_INITLIST,      /* initializer list (int a[] = {1, 2}; ) */
	T_GRAMMAR_TREE,  /* the root of the grammar subtree */
	T_TOKEN_TREE,    /* the root of the token subtree */
	T_TOKEN_STR,     /* a non-comment token in the token tree */
	T_TOKEN_PPDIR,   /* a preprocessor-directive token (#) */
	T_GENTREX,       /* subtree for all gentrex script results */
	T_GENTREX_RESULT,/* results of a script */
	T_GENTREX_MAP,   /* results of a script */
	T_SYMLINK,       /* name is an uid to a node to link to */
	T_INVALID        /* matches with the NULL in the name list */
} c99c_node_type_t;

extern const char *c99c_node_type_names[];

struct c99c_node_s {
	c99c_node_type_t type;
	int uid;

	char *name;         /* name field - textual payload for some nodes */
	c99c_loc_t loc;     /* location info */
	int print_data;     /* random data for the output layer */

	/* flags */
	int redefinable;    /* the type can be redefined without an error */

	c99c_node_t *next;
	c99c_node_t *children;

	/* added in post-processing */
	c99c_node_t *parent;

	/* temporary fields for building the tree */
	c99c_loc_t term_loc;          /* location info for the terminator token (';' usally) : not part of the the current node */

	/* for user code to store data */
	void *user_data;
};

/* builds/frees type hash */
void c99c_tree_init(c99tree_cc_t *ctx);
void c99c_tree_uninit(c99tree_cc_t *ctx);


/* Look up type by name */
c99c_node_type_t C99Ntype(c99tree_cc_t *ctx, const char *typenam);

#ifdef PARSER_NALLOC_DEBUG
# define YLINEARG   , int yline
# define YLINEPARM  , yline
#else
# define YLINEARG
# define YLINEPARM
#endif

/* Allocate a new tree node - ->next is empty, line is filled in from global lineno */
c99c_node_t *C99Nalloc(c99tree_cc_t *ctx, c99c_node_type_t type, char *name, c99c_node_t *child YLINEARG);

/* Same as Nalloc, set start and/or end location */
c99c_node_t *C99Nalloc_wloc(c99tree_cc_t *ctx, c99c_node_type_t type, char *name, c99c_node_t *child, const c99c_loc_t *start, const c99c_loc_t *end YLINEARG);

/* Same as Nalloc, but take a token (with start location) instead of a name */
c99c_node_t *C99Nalloc_tok(c99tree_cc_t *ctx, c99c_node_type_t type, c99c_strloc_t *token, c99c_node_t *child YLINEARG);

/* Append src node to the last ->next starting from dest - returns dest */
c99c_node_t *C99Nappend(c99tree_cc_t *ctx, c99c_node_t *dest, c99c_node_t *src);

/* unlink a subtree (rooted at node) from under its parent making it a standalone tree */
void C99Nunlink(c99tree_cc_t *ctx, c99c_node_t *node);

/* unlink a subtree (rooted at oldn) from under its parent making it a standalone tree
   and insert newn in the same place (if newn is not NULL) */
void C99Nreplace(c99tree_cc_t *ctx, c99c_node_t *oldn, c99c_node_t *newn);

/* Insert newn so that after->next is newn */
void C99Ninsert_after(c99tree_cc_t *ctx, c99c_node_t *after, c99c_node_t *newn);

/* Create operator node with 2 operands; 2nd operand may be NULL */
c99c_node_t *C99Nop(c99tree_cc_t *ctx, const char *op, c99c_node_t *o1, c99c_node_t *o2, c99c_loc_t *start, c99c_loc_t *end YLINEARG);
c99c_node_t *C99Nop_tok(c99tree_cc_t *ctx, const c99c_strloc_t *op, c99c_node_t *o1, c99c_node_t *o2 YLINEARG);

typedef enum invalidate_loc_e {
	INV_LOC_NONE,
	INV_LOC_ALL,
	INV_LOC_ALL_BUT_ID
} invalidate_loc_t;

/* Clone a subtree recursively */
c99c_node_t *C99Nclone(c99tree_cc_t *ctx, const c99c_node_t *node, invalidate_loc_t invalidate_loc YLINEARG);

/* Free a subtree recurisively */
void C99Nfree(c99tree_cc_t *ctx, c99c_node_t *node);

/* Print node and all ->next nodes recursively, prefix lines with prefix. If
   local_only is non-zero, print nodes found in the main file only, not the
   nodes from includes
   nprint_ctx should point to a void * that is NULL before the call; the value
   after the call should be ignored.
   */
void C99Nprint(c99tree_cc_t *ctx, void **nprint_ctx, const char *prefix, c99c_node_t *node, int local_only);

/* return the last node on the ->next list of node */
c99c_node_t *C99Nlast(c99tree_cc_t *ctx, c99c_node_t *node);

/* set dest loc to span between the leftmost and rightmost item of subtree child */
void C99Nloc_cpy_subtree(c99tree_cc_t *ctx, c99c_node_t *dest, c99c_node_t *child);

/* token tree: append token node */
c99c_node_t *c99c_tt_append(c99tree_cc_t *ctx, c99c_node_type_t type, char *text YLINEARG);

/* Append src node as dest node's last child
   Make copy of dest_ and src_ as they may be calls */
#define C99Nchappend(ctx, dest__, src__)	\
	do { \
		c99c_node_t *_dest_, *_src_; \
		_dest_ = dest__; \
		_src_ = src__; \
		if ((_dest_)->children != NULL) \
			C99Nappend(ctx, (_dest_)->children, (_src_)); \
		else \
			(_dest_)->children = _src_; \
	} while(0)


/* Insert src node as dest node's first child
   Make copy of dest_ and src_ as they may be calls */
#define C99Nchinsert(ctx, dest_, src_)	\
	do { \
		c99c_node_t *_dest_, *_src_; \
		_dest_ = dest_; \
		_src_ = src_; \
		if ((_dest_)->children != NULL) {\
			if (_src_ != NULL) { \
				C99Nappend(ctx, _src_, (_dest_)->children); \
				(_dest_)->children = _src_; \
			} \
		} \
		else \
			(_dest_)->children = _src_; \
	} while(0)


#define C99NTYinsert(ctx, dest_, src_) \
	do { \
		c99c_node_t *_dest_, *_src_; \
		_dest_ = dest_; \
		_src_ = src_; \
		if ((_dest_->children == NULL) || (_dest_->children->type != T_TYPE)) {\
			c99c_node_t *n__; \
			n__ = C99Nalloc(ctx, T_TYPE, NULL, _dest_->children); \
			if (_src_ != NULL) \
				C99Nloc_cpy(ctx, n__, _src_); \
			_dest_->children = n__; \
		} \
		_src_->next = _dest_->children->children; \
		_dest_->children->children = _src_; \
	} while(0)

#ifdef PARSER_NALLOC_DEBUG
# ifndef YLINEDEF
#  define YLINEDEF -1
# endif
# ifndef LLINEDEF
#  define LLINEDEF -1
# endif
# define C99Nalloc(ctx, type, name, child)                       C99Nalloc(ctx, type, name, child, YLINEDEF)
# define C99Nalloc_tok(ctx, type, tok, child)                    C99Nalloc_tok(ctx, type, tok, child, YLINEDEF)
# define C99Nclone(ctx, node, ivl)                               C99Nclone(ctx, node, ivl, YLINEDEF)
# define C99Nop(ctx, op, o1, o2, s, e)                           C99Nop(ctx, op, o1, o2, s, e, YLINEDEF)
# define C99Nop_tok(ctx, op, o1, o2)                             C99Nop_tok(ctx, op, o1, o2, YLINEDEF)
# define C99Nalloc_wloc(ctx, type, name, child, start, end)      C99Nalloc_wloc(ctx, type, name, child, start, end, YLINEDEF)
# define c99c_tt_append(ctx, type, text)                              c99c_tt_append(ctx, type, text, LLINEDEF)
#endif

#define attrlist_is_post(node)  ((node->name != NULL) && (node->name[1] == 'o'))
#define attrlist_is_pre(node)  ((node->name != NULL) && (node->name[1] == 'r'))


/* postprocess the tree: fill in the parent field */
void c99c_tree_postproc(c99c_node_t *root);

/* rename node and it's right siblings to name */
void c99c_name_all_siblings(c99tree_cc_t *ctx, c99c_node_t *node, const char *name);

#endif
