/*
    c99tree - input layer, xml parser using libexpat
    Copyright (C) 2010-2015  Tibor Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <expat.h>
#include "c99tree-cc.h"
#include "tree_load.h"
#include "escaping.h"
#include "scope.h"
#include "tree_load_uid.h"

#define BUFFLEN 1024

typedef struct xml_ctx_s {
	c99c_node_t *root;
	c99c_node_t *last[LOADER_MAXLEVEL];
	c99c_uid2node_t *uid2node;
	c99tree_cc_t *tree_ctx;
	int level;
} xml_ctx_t;

static void xml_start(void *user_data, const XML_Char *type, const XML_Char **atts)
{
	c99c_node_t *n;
	const char **a;
	const char *name = NULL, *loc = NULL, *twin_parent = NULL, *uid = NULL;
	xml_ctx_t *ctx = user_data;
	c99c_loc_t locbuf;

	if (strcmp(type, "FILE") == 0) {
		for(a = atts; *a != NULL; a += 2)
			if (strcmp(a[0], "location_format") == 0) ctx->tree_ctx->loc_format = c99c_unescape_dup_xml(a[1]);
		return;
	}

	for(a = (const char **)atts; *a != NULL; a += 2) {
		if (strcmp(a[0], "name") == 0) name = a[1];
		else if (strcmp(a[0], "location") == 0) loc = a[1];
		else if (strcmp(a[0], "twin_parent") == 0) twin_parent = a[1];
		else if (strcmp(a[0], "uid") == 0) uid = a[1];
	}
	ctx->tree_ctx->loc_current = c99c_loc_scan(ctx->tree_ctx, loc == NULL ? "" : loc, &locbuf);


	n = C99Nalloc(ctx->tree_ctx, C99Ntype(ctx->tree_ctx, type), name == NULL ? NULL : c99c_unescape_dup_xml(name), NULL);

	/* bind new node */
	if (ctx->level == 0) {
		if (ctx->root == NULL)
			ctx->root = n;
		else
			C99Nappend(ctx->tree_ctx, ctx->root, n);
	}
	else {
		C99Nchappend(ctx->tree_ctx, ctx->last[ctx->level-1], n);
	}
	ctx->last[ctx->level] = n;

	n->uid = -1;

	if (uid != NULL) {
		int i;
		char *end;
		i = strtol(uid, &end, 10);
		if (*end == '\0') {
			n->uid = i;
			c99c_uid2node_save(ctx->uid2node, n);
		}
	}

	n->loc.twin_parent_uid = -1;
	n->loc.twin_parent = NULL;
	n->loc.is_local = 1;
	if (twin_parent != NULL) {
		int i;
		char *end;
		i = strtol(twin_parent, &end, 10);
		if (*end == '\0')
			n->loc.twin_parent_uid = i;
	}

	ctx->level++;
	assert(ctx->level < LOADER_MAXLEVEL);
}

static void xml_end(void *user_data, const XML_Char *name)
{
	xml_ctx_t *ctx = user_data;

	ctx->level--;
}


c99c_node_t *c99c_tree_load_xml(c99tree_cc_t *ctx, const char *filename)
{
	char buff[BUFFLEN];
	int len;
	void *parser;
	int fd;
	char *old_loc_format;
	xml_ctx_t xml_ctx;

	old_loc_format = ctx->loc_format;
	ctx->loc_format = NULL;

	memset(&xml_ctx, 0, sizeof(xml_ctx));
	xml_ctx.tree_ctx = ctx;

	if (filename != NULL) {
		fd = open(filename, O_RDONLY);
	}
	else {
		filename = "<stdin>";
		fd = 0;
	}

	xml_ctx.uid2node = c99c_uid2node_alloc();
	parser = XML_ParserCreate(NULL);
	assert(parser != NULL);

	xml_ctx.level = 0;
	XML_SetUserData(parser, &xml_ctx);
  XML_SetElementHandler(parser, xml_start, xml_end);

	for(;;) {
		len = read(0, buff, BUFFLEN);
		if (len <= 0)
			break;
		if (XML_Parse(parser, buff, len, len < BUFFLEN) == XML_STATUS_ERROR) {
			fprintf(stderr, "Error loading xml: %s:%lu\n",
				XML_ErrorString(XML_GetErrorCode(parser)),
				XML_GetCurrentLineNumber(parser));
			abort();
		}
	}
	XML_ParserFree(parser);

	if (fd != 0)
		close(fd);

	if (ctx->loc_format != NULL)
		free(ctx->loc_format);

	ctx->loc_format = old_loc_format;
	ctx->loc_current = NULL;
	c99c_uid2node_resolve(xml_ctx.uid2node, xml_ctx.root);
	c99c_uid2node_free(xml_ctx.uid2node);
	return xml_ctx.root;
}

